# coding=utf-8
"""
Functions intended to be used in machine learning problems.

Available Functions
-------------------
[Public]

windowing
    Function to window the input signal, that return the windowed signal.

zero_crossing_rate
    Get the zero-crossing rate of the input signal.

features_extraction
    Function to extract features given in functions, that returns an array with the set of features
    for each time window.

Observations/Comments
---------------------
None

/\
"""
import numpy as np


def windowing(signal, sampling_rate=1000, time_window=.25, overlap=0):
    """
    -----
    Brief
    -----
    Function to window the input signal, that return the windowed signal.

    -----------
    Description
    -----------
    Machine learning pipelines usually use features to represent each sample of the input signals.
    In order to extract those features, it is required to window the signal in time intervals,
    in the case of time series, that may be or may not be of the same duration with or without overlap.

    This function allows to window the input signal in similar time windows with or without overlap.

    ----------
    Parameters
    ----------
    signal : list or numpy.array
        Signal to be windowed.
    sampling_rate : int
        Sampling rate of the input signal.
    time_window : float or int
        Time in seconds of each window.
	overlap : float
		Percentage of overlap relative to the time window.

    Returns
    -------
        Windowed signal

    """
    if  overlap < 0 or overlap >= 1:
        raise ValueError("overlap should be a float in [0, 1)")
    over = 1 - overlap
    
    app_window = int(sampling_rate * time_window)

    num_windows = int(len(signal) / (app_window*over))
    signal_windows = np.zeros(shape=(num_windows, app_window))

    for i, win in enumerate(range(0, len(signal), int(over*app_window))):
        if win + app_window < len(signal) and i < num_windows:
            signal_windows[i] = signal[win:win + app_window]
    return signal_windows


def zero_crossing_rate(signal):
    """
    -----
    Brief
    -----
    Get the zero-crossing rate of the input signal.

    -----------
    Description
    -----------
    Zero-crossing rate is a widely used feature used to represent a signal. This feature gives the number of times
    a signal cross the zero, which means, the number of times it changes polarity.

    This function returns an int indicating the number of times the input signal crosses the zero value.

    ----------
    Parameters
    ----------

    signal : list or numpy.array
        Input signal.

    Returns
    -------
        Number of times the input signal crosses the zero value.
    """
    return (np.diff(np.sign(signal)) != 0).sum()


def features_extraction(windowed_signal, functions):
    """
    -----
    Brief
    -----
    Function to extract features given in functions, that returns an array with the set of features
    for each time window.

    -----------
    Description
    -----------
    Machine learning pipelines usually use features to represent each sample of the input signals.
    Those features should be relevant in the sense that they should be useful to distinguish between the classes
    of each specific problem, and not redundant, as the usage of redundant features usually improves the complexity
    of the problem without improving its results.

    This function allows to extract features from the various windows provided in windowed_signal.

    ----------
    Parameters
    ----------
    windowed_signal: list or numpy.array
        Input windowed signal (if your signal is not windowed, you can use the function windowing of this module).
    functions: list
        List of functions that will be applied to each window. For example: [numpy.mean, numpy.std]

    Returns
    -------
        Array of features with shape (n_windows, n_features).
    """
    features = np.zeros(shape=(windowed_signal.shape[0], len(functions)))

    for i, window in enumerate(windowed_signal):
        fea = []
        for f in functions:
            fea.append(f(window))
        features[i] = fea

    return features



# coding=utf-8
"""
Processing capabilities that are more general than the remaining modules categories.

Available Functions
-------------------
[Public]

poincare
    With this function the user can easily generate a Poincaré Plot (Heart rate variability
    analysis).

smooth
    Function intended to smooth a signal through the application of the convolution operation
    between a moving average window (the signal is segmented into multiple parts) and a fixed window
    (from one of the predefined formats 'hanning', 'blackman'...) in order to a weigh be attributed
    to each sample inside the moving window.

plotfft
    This functions computes the Fast Fourier Transform of a signal, returning the frequency and magnitude values.

lowpass
    For a given signal s rejects (attenuates) the frequencies higher than the cutoff frequency f and passes the
    frequencies lower than that value by applying a Butterworth digital filter.

highpass
    For a given signal s rejects (attenuates) the frequencies lower then the cutoff frequency f and passes the
    frequencies higher than that value by applying a Butterworth digital filter.

bandstop
    For a given signal s rejects (attenuates) the frequencies within a certain range (between f1 and f2) and passes the
    frequencies outside that range by applying a Butterworth digital filter.

bandpass
    For a given signal s passes the frequencies within a certain range (between f1 and f2) and rejects (attenuates) the
    frequencies outside that range by applying a Butterworth digital filter.

Observations/Comments
---------------------
None

/\
"""

import numpy
from scipy.signal import filtfilt, butter, lfilter
from .detect import tachogram
from .visualise import plot


def poincare(data, sample_rate, signal=False, in_seconds=False):
    """
    -----
    Brief
    -----
    Function for generation of Poincaré Plot (Heart rate variability analysis).

    -----------
    Description
    -----------
    ECG signals measure the electric potential in the heart of the subject. In normal conditions, it is expeted that the
    the electric potential to be similar in different heartbeats and that the rhythm of those heartbeats to be
    maintained if all the conditions are maintained. Thus, by plotting the current RR interval against the previous one,
    it is expected that the values to be maintained. Poincaré plot, is this representation, which allows to analyse the
    heart rate variability.

    This function returns the x and y axis of a Poincaré plot and also the standard deviations of the more
    representative directions of the data points.

    ----------
    Parameters
    ----------
    data : list
        ECG signal or R peak list. When the input is a raw signal the input flag signal should be
        True.

    sample_rate : int
        Sampling frequency.

    signal : boolean
        If True, then the data argument contains the set of the ECG acquired samples.

    in_seconds : boolean
        If the R peaks list defined as the input argument "data" contains the sample numbers where
        the R peaks occur,
        then in_seconds needs to be True.

    Returns
    -------
    out : list, list, float, float
        Poincaré plot x axis and y axis, respectively. Additionally it will be returned SD1 and SD2
        parameters.

    """

    # Generation of tachogram.
    tachogram_data = tachogram(data, sample_rate, signal=signal, in_seconds=in_seconds,
                               out_seconds=True)[0]

    # Poincaré Plot (x and y axis).
    x_axis = tachogram_data[:-1]
    y_axis = tachogram_data[1:]

    # Poincaré Parameters.
    tachogram_diff = numpy.diff(tachogram_data)
    sdsd = numpy.std(tachogram_diff)
    sdnn = numpy.std(tachogram_data)

    sd1 = numpy.sqrt(0.5 * numpy.power(sdsd, 2))
    sd2 = numpy.sqrt(2 * numpy.power(sdnn, 2) - numpy.power(sd1, 2))

    return x_axis, y_axis, sd1, sd2


def smooth(input_signal, window_len=10, window='hanning'):
    """
    -----
    Brief
    -----
    Smooth the data using a window with requested size.

    -----------
    Description
    -----------
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal (with the window size) in both ends so that
    transient parts are minimized in the beginning and end part of the output signal.

    The results of the application of this functions is analogous to the application of a mean filter in image
    processing. The results is the smoothed input_signal.

    ----------
    Parameters
    ----------
    input_signal: array-like
        the input signal
    window_len: int
        the dimension of the smoothing window. the default is 10.
    window: string.
        the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'.
        flat window will produce a moving average smoothing. the default is 'hanning'.

    Returns
    -------
    out :   signal_filt: array-like the smoothed signal.

    @example:
                time = linspace(-2,2,0.1)
                input_signal = sin(t)+randn(len(t))*0.1
                signal_filt = smooth(x)


    @see also:  numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman,
                numpy.convolve, scipy.signal.lfilter


    @todo: the window parameter could be the window itself if an array instead
    of a string

    @bug: if window_len is equal to the size of the signal the returning
    signal is smaller.
    """

    if input_signal.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if input_signal.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return input_signal

    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("""Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'""")

    sig = numpy.r_[2 * input_signal[0] - input_signal[window_len:0:-1],
                   input_signal, 2 * input_signal[-1] - input_signal[-2:-window_len-2:-1]]

    if window == 'flat':  # moving average
        win = numpy.ones(window_len, 'd')
    else:
        win = eval('numpy.' + window + '(window_len)')

    sig_conv = numpy.convolve(win / win.sum(), sig, mode='same')

    return sig_conv[window_len: -window_len]


def plotfft(s, fmax, doplot=False):
    """
    -----
    Brief
    -----
    This functions computes the Fast Fourier Transform of a signal, returning the frequency and magnitude values.

    -----------
    Description
    -----------
    Fast Fourier Transform (FFT) is a method to computationally calculate the Fourier Transform of discrete finite
    signals. This transform converts the time domain signal into a frequency domain signal by abdicating the temporal
    dimension.

    This function computes the FFT of the input signal and returns the frequency and respective amplitude values.

    ----------
    Parameters
    ----------
    s: array-like
      the input signal.
    fmax: int
      the sampling frequency.
    doplot: boolean
      a variable to indicate whether the plot is done or not.

    Returns
    -------
    f: array-like
      the frequency values (xx axis)
    fs: array-like
      the amplitude of the frequency values (yy axis)
    """

    fs = abs(numpy.fft.fft(s))
    f = numpy.linspace(0, fmax / 2, len(s) / 2)
    if doplot:
        plot(list(f[1:int(len(s) / 2)]), list(fs[1:int(len(s) / 2)]))
    return f[1:int(len(s) / 2)].copy(), fs[1:int(len(s) / 2)].copy()


def lowpass(s, f, order=2, fs=1000.0, use_filtfilt=False):
    """
    -----
    Brief
    -----
    For a given signal s rejects (attenuates) the frequencies higher than the cutoff frequency f and passes the
    frequencies lower than that value by applying a Butterworth digital filter.

    -----------
    Description
    -----------
    Signals may have frequency components of multiple bands. If our interest is to have an idea about the behaviour
    of low frequency bands, we should apply a low pass filter, which would attenuate the higher frequencies of the
    signal. The degree of attenuation is controlled by the parameter "order", that as it increases, allows to better
    attenuate frequencies closer to the cutoff frequency. Notwithstanding, the higher the order, the higher the
    computational complexity and the higher the instability of the filter that may compromise the results.

    This function allows to apply a low pass Butterworth digital filter and returns the filtered signal.

    ----------
    Parameters
    ----------
    s: array-like
        signal
    f: int
        the cutoff frequency
    order: int
        Butterworth filter order
    fs: float
        sampling frequency
    use_filtfilt: boolean
        If True, the signal will be filtered once forward and then backwards. The result will have zero phase and twice
        the order chosen.

    Returns
    -------
    signal: array-like
        filtered signal

    """
    b, a = butter(order, f / (fs/2))

    if use_filtfilt:
        return filtfilt(b, a, s)

    return lfilter(b, a, s)


def highpass(s, f, order=2, fs=1000.0, use_filtfilt=False):
    """
    -----
    Brief
    -----
    For a given signal s rejects (attenuates) the frequencies lower then the cutoff frequency f and passes the
    frequencies higher than that value by applying a Butterworth digital filter.

    -----------
    Description
    -----------
    Signals may have frequency components of multiple bands. If our interest is to have an idea about the behaviour
    of high frequency bands, we should apply a high pass filter, which would attenuate the lower frequencies of the
    signal. The degree of attenuation is controlled by the parameter "order", that as it increases, allows to better
    attenuate frequencies closer to the cutoff frequency. Notwithstanding, the higher the order, the higher the
    computational complexity and the higher the instability of the filter that may compromise the results.

    This function allows to apply a high pass Butterworth digital filter and returns the filtered signal.

    ----------
    Parameters
    ----------
    s: array-like
        signal
    f: int
        the cutoff frequency
    order: int
        Butterworth filter order
    fs: float
        sampling frequency
    use_filtfilt: boolean
        If True, the signal will be filtered once forward and then backwards. The result will have zero phase and twice
        the order chosen.

    Returns
    -------
    signal: array-like
        filtered signal

    """

    b, a = butter(order, f * 2 / (fs/2), btype='highpass')
    if use_filtfilt:
        return filtfilt(b, a, s)

    return lfilter(b, a, s)


def bandstop(s, f1, f2, order=2, fs=1000.0, use_filtfilt=False):
    """
    -----
    Brief
    -----
    For a given signal s rejects (attenuates) the frequencies within a certain range (between f1 and f2) and passes the
    frequencies outside that range by applying a Butterworth digital filter.

    -----------
    Description
    -----------
    Signals may have frequency components of multiple bands. If our interest is to have an idea about the behaviour
    of all frequency bands except a specific band, we should apply a band stop filter, which would attenuate the
    frequencies of that band of the signal. The degree of attenuation is controlled by the parameter "order", that as it
    increases, allows to better attenuate frequencies closer to the cutoff frequencies. Notwithstanding, the higher the
    order, the higher the computational complexity and the higher the instability of the filter which may compromise the
    results.

    This function allows to apply a band stop Butterworth digital filter and returns the filtered signal.

    ----------
    Parameters
    ----------
    s: array-like
        signal
    f1: int
        the lower cutoff frequency
    f2: int
        the upper cutoff frequency
    order: int
        Butterworth filter order
    fs: float
        sampling frequency
    use_filtfilt: boolean
        If True, the signal will be filtered once forward and then backwards. The result will have zero phase and twice
        the order chosen.

    Returns
    -------
    signal: array-like
        filtered signal

    """
    b, a = butter(order, [f1 * 2 / fs, f2 * 2 / fs], btype='bandstop')
    if use_filtfilt:
        return filtfilt(b, a, s)
    return lfilter(b, a, s)


def bandpass(s, f1, f2, order=2, fs=1000.0, use_filtfilt=False):
    """
    -----
    Brief
    -----
    For a given signal s passes the frequencies within a certain range (between f1 and f2) and rejects (attenuates) the
    frequencies outside that range by applying a Butterworth digital filter.

    -----------
    Description
    -----------
    Signals may have frequency components of multiple bands. If our interest is to have an idea about the behaviour
    of a specific frequency band, we should apply a band pass filter, which would attenuate all the remaining
    frequencies of the signal. The degree of attenuation is controlled by the parameter "order", that as it increases,
    allows to better attenuate frequencies closer to the cutoff frequency. Notwithstanding, the higher the order, the
    higher the computational complexity and the higher the instability of the filter that may compromise the results.

    This function allows to apply a band pass Butterworth digital filter and returns the filtered signal.

    ----------
    Parameters
    ----------
    s: array-like
        signal
    f1: int
        the lower cutoff frequency
    f2: int
        the upper cutoff frequency
    order: int
        Butterworth filter order
    fs: float
        sampling frequency
    use_filtfilt: boolean
        If True, the signal will be filtered once forward and then backwards. The result will have zero phase and twice
        the order chosen.

    Returns
    -------
    signal: array-like
        filtered signal

    """
    b, a = butter(order, [f1 * 2 / fs, f2 * 2 / fs], btype='bandpass')

    if use_filtfilt:
        return filtfilt(b, a, s)

    return lfilter(b, a, s)


# 25/09/2018 18h58m :)



"""
OpenSignalsTools module intended to give the user the possibility of programmatically generate a
Notebook.

"""

import json
import os
import re
import shutil
import warnings
import importlib.util

import nbformat as nb
from .notebook_files.MainFiles.group_by_difficulty import STAR_TABLE_HEADER
from .notebook_files.MainFiles.group_by_tag import TAG_TABLE_HEADER

from .notebook_files.MainFiles.signal_samples import IMPORT_CODE, PLOTS_SIGNAL_SAMPLES

from .notebook_files.cell_content_strings import DESCRIPTION_GROUP_BY, DESCRIPTION_SIGNAL_SAMPLES, \
                                 HEADER_ALL_CATEGORIES, HEADER_MAIN_FILES, \
                                 DESCRIPTION_CATEGORY, HEADER_TAGS, SEPARATOR, AUX_CODE_MESSAGE, \
                                 JS_CODE_AUTO_PLAY, CSS_STYLE_CODE, FOOTER, HEADER, MD_EXAMPLES, \
                                 CODE_EXAMPLES, OPEN_IMAGE

SIGNAL_TYPE_LIST = ["emg", "ecg"]

NOTEBOOK_KEYS = {"Install": 13, "Connect": 14, "Record": 2, "Load": 1, "Visualise": 3, "Pre-Process": 4, "Detect": 5,
                 "Extract": 6, "Train_And_Classify": 7, "Understand": 8, "Evaluate": 12,
                 "Other": 15, "MainFiles": 0}

# ==================================================================================================
# ======================================= notebook Class ===========================================
# ==================================================================================================


class notebook:
    def __init__(self, notebook_type=None, notebook_title="Notebook Title", tags="tags",
                 difficulty_stars=1, notebook_description="Notebook Description",
                 dict_by_difficulty=None, dict_by_tag=None, notebook_file=None):
        """
        -----
        Brief
        -----
        Class constructor that generates a new Notebook template, taking into account the
        specified 'notebook_type'.

        -----------
        Description
        -----------
        OpenSignals notebooks have specific templates according to different types. They possess specific headers,
        footers and general arrangement and features, such as, the way of presenting the difficulty level and keywords.
        All of these elements need to be implemented using HTML and CSS in order to use all functionalities of the
        jupyter notebook environment.
        That environment allows to easily and clearly present and compile Python code as well as information in markdown
        cells, and has been widely used in data mining applications.

        This class allows to generate a jupyter notebook template in order to facilitate content creation without the
        need to worry about the design.

        ----------
        Parameters
        ----------
        notebook_type : str
            Notebook type: - "Main_Files_Signal_Samples"
                           - "Main_Files_By_Category"
                           - "Main_Files_By_Difficulty"
                           - "Main_Files_By_Tag"
                           - "Main_Files_By_Signal_Type"
                           - "Acquire"
                           - "Open"
                           - "Visualise"
                           - "Process"
                           - "Detect"
                           - "Extract"
                           - "Train_And_Classify"
                           - "Explain"

        notebook_title : None or str
            The Notebook title should only be defined when 'notebook_type' is:
           - "Acquire"
           - "Open"
           - "Visualise"
           - "Process"
           - "Detect"
           - "Extract"
           - "Train_And_Classify"
           - "Explain"

        tags : str
            Sequence of tags that characterize the Notebook.

        difficulty_stars : int
            This input defines the difficulty level of the Notebook instructions.

        notebook_description : str
            An introductory text to present the Notebook and involve the reader.

        dict_by_difficulty : dict
            Global Dictionary that groups Notebooks names/files by difficulty level.

        dict_by_tag : dict
            Dictionary where each key is a tag and the respective value will be a list containing the
            Notebooks (title and filename) that include this tag.

        notebook_file : str
            Notebook filename.
        """

        # ========================= Initialisation of Object variables =============================
        self.difficulty_stars = difficulty_stars
        self.tags = tags
        self.notebook_title = notebook_title

        # =============================== Creation of a Notebook ===================================
        self.notebook = nb.v4.new_notebook()

        if "Main" in notebook_type:
            warnings.warn("The arguments 'tags' and 'difficulty_stars' does not have effect for "
                          "the " + notebook_type + " notebook type !")

        if notebook_type in ["Load", "Record", "Visualise", "Pre-Process", "Detect",
                             "Extract", "Train_And_Classify", "Understand", "Evaluate", "Other", "Install", "Connect"]:
            self.notebook_type = notebook_type
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_notebook_header(self.notebook, notebook_type, notebook_title, tags,
                                      difficulty_stars, notebook_description)

        elif notebook_type == "Main_Files_By_Category":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_notebooks_by_category(self.notebook, dict_by_tag)
            if os.path.exists("../biosignalsnotebooks_notebooks"):
                _generate_github_readme(self.notebook, dict_by_tag)

        elif notebook_type == "Main_Files_Signal_Samples":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_SIGNAL_SAMPLES)
            _generate_signal_samples_body(self.notebook)

        elif notebook_type == "Main_Files_By_Difficulty":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_GROUP_BY)
            #opensignals_hierarchy()
            _generate_notebook_by_difficulty_body(self.notebook, dict_by_difficulty)

        elif notebook_type == "Main_Files_By_Tag":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_GROUP_BY)
            # opensignals_hierarchy()
            _generate_notebook_by_tag_body(self.notebook, dict_by_tag)

        elif notebook_type == "Main_Files_By_Signal_Type":
            self.notebook_type = "MainFiles"
            _generate_header(self.notebook, self.notebook_type, notebook_file)
            _generate_main_files_header(self.notebook, notebook_title, DESCRIPTION_GROUP_BY)
            # opensignals_hierarchy()
            _generate_notebook_by_signal_type_body(self.notebook, dict_by_tag)

    def write_to_file(self, path, filename, footer=True, group_by=False):
        """
        -----
        Brief
        -----
        Class method responsible for generating a file containing the notebook object data.

        -----------
        Description
        -----------
        biosignalsnotebooks allows to create jupyter notebook objects that includes all of the information required to
        generate a jupyter notebook file.

        This function write a jupyter notebook file from a jupyter notebook object.

        ----------
        Parameters
        ----------
        path : str
            OpenSignalsTools Root folder path (where the notebook will be stored).

        filename : str
            Defines the name of the notebook file.

        footer : bool
            Flag that defines when the footer needs to be included in the Notebook.

        group_by : bool
            If True then the group by pages that compose the biosignalsnotebooks project will be
            automatically generated (grouping Notebooks by difficulty, tags ...)
        """

        # =============================== Storage of Filename ======================================
        self.filename = filename

        # ======================== Inclusion of Footer in the Notebook =============================
        if footer is True:
            _generate_footer(self.notebook, self.notebook_type)

        # ========== Code segment for application of the OpenSignalsTools CSS style ===========
        self.notebook["cells"].append(nb.v4.new_markdown_cell(AUX_CODE_MESSAGE,
                                                              **{"metadata":
                                                                     {"tags": ["hide_mark",
                                                                               "aux"]}}))
        self.notebook["cells"].append(nb.v4.new_code_cell(CSS_STYLE_CODE,
                                                          **{"metadata":
                                                                 {"tags": ["hide_both"]}}))
        self.notebook["cells"].append(nb.v4.new_code_cell(JS_CODE_AUTO_PLAY,
                                                          **{"metadata":
                                                                 {"tags": ["hide_both"]}}))

        full_path = path + "/Categories/" + self.notebook_type + "/" + filename + ".ipynb"

        # Check if file exists.
        if not os.path.exists(full_path) or self.notebook_type == "MainFiles":
            nb.write(self.notebook, full_path)
        else:
            raise RuntimeError("The specified filename already exists. The Notebook creation was interrupted !")

        # ========================== Run Notebook Code Instructions ================================
        os.system("jupyter nbconvert --execute --inplace --ExecutePreprocessor.timeout=-1 " +
                  full_path)
        os.system("jupyter trust " + full_path)

        # ==== Automatic generation of pages that group the Notebooks by different criteria ========
        if group_by is True:
            _generate_group_by_pages(path)

    def add_markdown_cell(self, content, tags=None):
        """
        -----
        Brief
        -----
        Class method responsible for adding a markdown cell with content 'content' to the
        Notebook object.

        -----------
        Description
        -----------
        Jupyter notebooks allow to use multiple format cells. One of those formats is the markdown cell, which allows
        to write informational content, write HTML and CSS code and structure the overall content of the notebook.

        This function allows to programmatically create a markdown cell with the content specified in the respective
        input.

        ----------
        Parameters
        ----------
        content : str
            Text/HTML code/... to include in the markdown cell (triple quote for multiline text).

        tags : list
            A list of tags to include in the markdown cell metadata.
        """
        self.notebook["cells"].append(nb.v4.new_markdown_cell(content, **{"metadata":
                                                                          {"tags": tags}}))

    def add_code_cell(self, content, tags=None):
        """
        -----
        Brief
        -----
        Class method responsible for adding a code cell with content 'content' to the
        Notebook object.

        -----------
        Description
        -----------
        Jupyter notebooks allow to use multiple format cells. One of those formats is the code cell, which allows
        to write Python code that can be ran in the notebooks.

        This function allows to programmatically create a code cell with the content specified in the respective
        input.

        ----------
        Parameters
        ----------
        content : str
            Code in a string format to include in the cell (triple quote for multiline
            text).

        tags : list
            A list of tags to include in the code cell metadata.
        """
        self.notebook["cells"].append(nb.v4.new_code_cell(content, **{"metadata":
                                                                      {"tags": tags}}))

# ==================================================================================================
# ========================== Generate OpenSignalTools File Hierarchy ===============================
# ==================================================================================================


def opensignals_hierarchy(root=None, update=False, clone=False):
    """
    -----
    Brief
    -----
    Function that generates the OpenSignalsTools Notebooks File Hierarchy programatically.

    -----------
    Description
    -----------
    OpenSignalsTools Notebooks folder obey to a predefined hierarchy that allows to run the code of online available
    notebooks and using the sample files, figures and CSS files in an easy way.

    This function generates the folder hierarchy programatically and automatically.

    ----------
    Parameters
    ----------
    root : None or str
        The file path where the OpenSignalsTools Environment will be stored.

    update : bool
        If True the old files will be replaced by the new ones.

    clone : bool
        If True then all the available Notebooks will be stored in the users computer.
        If False only the folder hierarchy of OpenSignalsTools will be generated, giving to the
        user a blank template for creating his own Notebook Environment.

    Returns
    -------
    out : str
        The root file path of OpenSignalsTools Environment is returned.
    """

    if root is None:
        root = os.getcwd()

    categories = list(NOTEBOOK_KEYS.keys())

    # ============================ Creation of the main directory ==================================
    current_dir = root + "/biosignalsnotebooks_environment"
    if not os.path.isdir(current_dir):
        os.makedirs(current_dir)

    # ================== Copy of 'images' 'styles' and 'signal_samples' folders ====================
    package_path = os.path.abspath(__file__).split(os.path.basename(__file__))[0].replace("\\", "/")
    for var in ["images", "styles", "signal_samples"]:
        if not os.path.isdir(root + "/biosignalsnotebooks_environment/" + var):
            src = (package_path + "notebook_files/osf_files/" + var).replace("\\", "/")
            destination = current_dir + "/" + var
            shutil.copytree(src, destination)
        elif update is True:
            shutil.rmtree(root + "/biosignalsnotebooks_environment/" + var)
            src = (package_path + "notebook_files/osf_files/" + var).replace("\\", "/")
            destination = current_dir + "/" + var
            shutil.copytree(src, destination)

    # =========================== Generation of 'Categories' folder ================================
    current_dir = root + "/biosignalsnotebooks_environment/Categories"
    if not os.path.isdir(current_dir):
        os.makedirs(current_dir)

    for category in categories:
        if not os.path.isdir(current_dir + "/" + category):
            os.makedirs(current_dir + "/" + category)
            if category == "MainFiles":
                # Copy aux folder.
                src = "../biosignalsnotebooks_notebooks/Categories/MainFiles/aux_files"
                shutil.copytree(src, current_dir + "/" + category + "/aux_files")

    # if clone is True:
    #     # Fill each folder inside "Categories" directory with the respective notebooks.
    #     # Each notebook will be created by a specific function.
    #     dir_path = root + "/notebook_code"
    #     list_of_code_dirs = os.listdir(dir_path)
    #     for folder in list_of_code_dirs:
    #         folder_path = root + "/notebook_code/" + folder
    #         if folder != "MainFiles" and folder != "__pycache__":
    #             list_of_code_files = os.listdir(folder_path)
    #             for file in list_of_code_files:
    #                 if file != "__pycache__":
    #                     spec = importlib.util.spec_from_file_location(file, folder_path +
    #                                                                   "/" + file)
    #                     foo = importlib.util.module_from_spec(spec)
    #                     spec.loader.exec_module(foo)
    #                     foo.run(root + "/biosignalsnotebooks_environment")
    #
    #     # Generation of biosignalsnotebooks environment main files.
    #     main_page = notebook("Main_Files_By_Category")
    #     main_page.write_to_file(root + "/biosignalsnotebooks_environment", "biosignalsnotebooks",
    #                             footer=False)
    #
    #     by_difficulty = notebook("Main_Files_By_Difficulty", "Notebooks Grouped by Difficulty", notebook_description=DESCRIPTION_GROUP_BY)
    #     by_difficulty.write_to_file(root + "/biosignalsnotebooks_environment", "by_diff",
    #                                 footer=False)
    #
    #     by_tags = notebook("Main_Files_By_Tag", "Notebooks Grouped by Tag Values",
    #                        notebook_description=DESCRIPTION_GROUP_BY)
    #     by_tags.write_to_file(root + "/biosignalsnotebooks_environment", "by_tag",
    #                           footer=False)
    #
    #     by_signal_type = notebook("Main_Files_By_Signal_Type", "Notebooks Grouped by Signal Type",
    #                               notebook_description=DESCRIPTION_GROUP_BY)
    #     by_signal_type.write_to_file(root + "/biosignalsnotebooks_environment",
    #                                  "by_signal_type", footer=False)
    #
    #     signal_samples = notebook("Main_Files_Signal_Samples", "Signal Samples Library",
    #                               notebook_description=DESCRIPTION_SIGNAL_SAMPLES)
    #     signal_samples.write_to_file(root + "/biosignalsnotebooks_environment",
    #                                  "signal_samples", footer=False)

    return root + "/biosignalsnotebooks_environment"

# ==================================================================================================
# ========================= Auxiliary Functions for Notebook Generation ============================
# ==================================================================================================


def _generate_notebook_header(notebook_object, notebook_type, notebook_title="Notebook Title",
                              tags="tags", difficulty_stars=1,
                              notebook_description="Notebook Description"):
    """
    Internal function that is used for generation of the generic notebooks header.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the header will be created.

    notebook_type : str
        Notebook type: - "Main_Files_Signal_Samples"
                       - "Main_Files_By_Category"
                       - "Main_Files_By_Difficulty"
                       - "Main_Files_By_Tag"
                       - "Acquire"
                       - "Open"
                       - "Visualise"
                       - "Process"
                       - "Detect"
                       - "Extract"
                       - "Train_And_Classify"
                       - "Explain"

    notebook_title : None or str
        The Notebook title should only be defined when 'notebook_type' is:
       - "Acquire"
       - "Open"
       - "Visualise"
       - "Process"
       - "Detect"
       - "Extract"
       - "Train_And_Classify"
       - "Explain"

    tags : str
        Sequence of tags that characterize the Notebook.

    difficulty_stars : int
        This input defines the difficulty level of the Notebook instructions.

    notebook_description : str
        An introductory text to present the Notebook and involve the reader.

    """

    # ============================= Creation of Header ====================================
    header_temp = HEADER_ALL_CATEGORIES.replace("header_image_color_i", "header_image_color_" +
                                                str(NOTEBOOK_KEYS[notebook_type]))
    header_temp = header_temp.replace("header_image_i", "header_image_" +
                                      str(NOTEBOOK_KEYS[notebook_type]))
    header_temp = header_temp.replace("Notebook Title", notebook_title)
    notebook_object["cells"].append(nb.v4.new_markdown_cell(header_temp, **{"metadata": {"tags": ["intro_info_title"]}}))

    # =============== Inclusion of the div with "Difficulty" and "Tags" ===================
    tags_and_diff = HEADER_TAGS.replace('<td class="shield_right" id="tags">tags</td>',
                                        '<td class="shield_right" id="tags">' + "&#9729;".join(tags)
                                        + '</td>')
    for star in range(1, 6):
        if star <= difficulty_stars:
            tags_and_diff = tags_and_diff.replace("fa fa-star " + str(star), "fa fa-star "
                                                                             "checked")
        else:
            tags_and_diff = tags_and_diff.replace("fa fa-star " + str(star), "fa fa-star")

    notebook_object["cells"].append(nb.v4.new_markdown_cell(tags_and_diff,
                                                            **{"metadata": {"tags": ["intro_info_tags"]}}))

    # ================= Insertion of the div reserved to the Notebook Description ==================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(notebook_description,
                                                            **{"metadata":
                                                                {"tags": ["test"]}}))
    notebook_object["cells"].append(nb.v4.new_markdown_cell(SEPARATOR))

    # ======================= Insertion of a blank Markdown and Code cell ==========================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(MD_EXAMPLES))
    notebook_object["cells"].append(nb.v4.new_code_cell(CODE_EXAMPLES))


def _generate_main_files_header(notebook_object, notebook_title="Notebook Title",
                                notebook_description="Notebook Description"):
    """
    Internal function that is used for generation of the 'MainFiles' notebooks header.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the header will be created.

    notebook_title : None or str
        Title of the Notebook.

    notebook_description : str
        An introductory text to present the Notebook and involve the reader.

    """

    # =================================== Creation of Header =======================================
    header_temp = HEADER_MAIN_FILES.replace("Notebook Title", notebook_title)
    notebook_object["cells"].append(nb.v4.new_markdown_cell(header_temp))

    # ================ Insertion of the div reserved to the Notebook Description ===================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(notebook_description,
                                                            **{"metadata":
                                                               {"tags": ["test"]}}))
    # notebook_object["cells"].append(nb.v4.new_markdown_cell(SEPARATOR))


def _generate_footer(notebook_object, notebook_type):
    """
    Internal function that is used for generation of the notebooks footer.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the footer will be created.

    notebook_type : str
        Notebook type: - "Main_Files_Signal_Samples"
                       - "Main_Files_By_Category"
                       - "Main_Files_By_Difficulty"
                       - "Main_Files_By_Tag"
                       - "Acquire"
                       - "Open"
                       - "Visualise"
                       - "Process"
                       - "Detect"
                       - "Extract"
                       - "Train_And_Classify"
                       - "Explain"

    """

    footer_aux = FOOTER
    if "Main_Files" in notebook_type:
        footer_aux = footer_aux.replace("../MainFiles/", "")

    # ================ Insertion of the div reserved to the Notebook Description ===================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(footer_aux,
                                                            **{"metadata":
                                                               {"tags": ["footer"]}}))


def _generate_header(notebook_object, notebook_type, notebook_file):
    """
    Internal function that is used for generation of the notebooks header.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the header will be created.

    notebook_file : str
        Notebook filename.

    notebook_type : str
        Notebook type: - "Main_Files_Signal_Samples"
                       - "Main_Files_By_Category"
                       - "Main_Files_By_Difficulty"
                       - "Main_Files_By_Tag"
                       - "Acquire"
                       - "Open"
                       - "Visualise"
                       - "Process"
                       - "Detect"
                       - "Extract"
                       - "Train_And_Classify"
                       - "Explain"
                       - "Other"
                       - "Install"
                       - "Connect"

    """

    header_aux = HEADER
    if notebook_file != None:
        header_aux = header_aux.replace("FILENAME", notebook_file.split(".")[0] + ".zip")
        header_aux = header_aux.replace("SOURCE", "https://mybinder.org/v2/gh/biosignalsplux/biosignalsnotebooks/mybinder_complete?filepath=biosignalsnotebooks_environment%2Fcategories%2F" + notebook_type + "%2F" + notebook_file + ".dwipynb")

    if "Main_Files" in notebook_type:
        header_aux = header_aux.replace("../MainFiles/", "")

    # ================ Insertion of the div reserved to the Notebook Description ===================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(header_aux,
                                                            **{"metadata":
                                                               {"tags": ["header"]}}))


def _generate_signal_samples_body(notebook_object):
    """
    Internal function that is used for the generation of the table that enumerates all signal samples and respective
    descriptions.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the signal samples table is generated.

    """

    # =========================== Code Cell for Importing Packages =================================
    notebook_object["cells"].append(nb.v4.new_code_cell(IMPORT_CODE,
                                                        **{"metadata":
                                                            {"tags": ["hide_both"]}}))

    # ========= Generation of a table that synthesises the information about each signal ===========
    signal_samples_dir_jupyter = "../../signal_samples"
    signal_samples_dir_project = os.path.abspath(__file__).split(os.path.basename(__file__))[0].replace("\\", "/") + "notebook_files/osf_files/signal_samples"
    list_of_files = os.listdir(signal_samples_dir_project)
    for file in list_of_files:
        if ".json" not in file and ".h5" in file:
            file_name = file.split(".")[0]
            with open(signal_samples_dir_project + "/" + file_name + '_info.json') as json_data:
                signal_info = json.load(json_data)

            info_table = "<table width='100%'>\n" \
                         "\t<tr>\n" \
                         "\t\t<td colspan='2' class='signal_samples_header'>" + file_name + \
                         "</td>\n" + "</tr>\n"
            info_keys = list(signal_info.keys())
            for key in info_keys:
                info_table += "\t<tr>\n" \
                              "\t\t<td class='signal_samples_info_keys'>" + key + "</td>\n" + \
                              "\t\t<td class='signal_samples_info_values'>" + signal_info[key] + \
                              "</td>\n" + \
                              "\t</tr>\n"

            info_table += "</table>"

            # ============== Insertion of the HTML table inside a markdown cell ===================
            notebook_object["cells"].append(nb.v4.new_markdown_cell(info_table))

            # ========================= Graphical Representation of Signals =======================
            path_str = '"' + (signal_samples_dir_jupyter + "\\").replace("\\", "/") + '"'
            plot_signal_samples_temp = re.sub(re.compile(r'\bsignal_samples_dir\b'),
                                              path_str, PLOTS_SIGNAL_SAMPLES)
            plot_signal_samples_temp = re.sub(re.compile(r'\bfile\b'),
                                              '"' + str(file) + '"',
                                              plot_signal_samples_temp)
            notebook_object["cells"].append(nb.v4.new_code_cell(plot_signal_samples_temp,
                                                                **{"metadata":
                                                                    {"tags": ["hide_in"]}}))


def _generate_notebook_by_difficulty_body(notebook_object, dict_by_difficulty):
    """
    Internal function that is used for generation of the page where notebooks are organized by
    difficulty level.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_difficulty : dict
        Global Dictionary that groups Notebooks names/files by difficulty level.

    """

    difficulty_keys = list(dict_by_difficulty.keys())
    difficulty_keys.sort()
    for difficulty in difficulty_keys:
        markdown_cell = STAR_TABLE_HEADER
        markdown_cell = _set_star_value(markdown_cell, int(difficulty))
        for notebook_file in dict_by_difficulty[str(difficulty)]:
            split_path = notebook_file.replace("\\", "/").split("/")
            notebook_type = split_path[-2]
            notebook_name = split_path[-1].split("&")[0]
            notebook_title = split_path[-1].split("&")[1]
            markdown_cell += "\t<tr>\n\t\t<td width='20%' class='header_image_color_" + \
                             str(NOTEBOOK_KEYS[notebook_type]) + "'><img " \
                             "src='../../images/icons/" + notebook_type.title() +\
                             ".png' width='15%'>\n\t\t</td>"
            markdown_cell += "\n\t\t<td width='60%' class='center_cell open_cell_light'>" + \
                             notebook_title + "\n\t\t</td>"
            markdown_cell += "\n\t\t<td width='20%' class='center_cell'>\n\t\t\t<a href='" \
                             "../" + notebook_type.title() + "/" + notebook_name + \
                             "'><div class='file_icon'></div></a>\n\t\t</td>\n\t</tr>"

        markdown_cell += "</table>"

        # ==================== Insertion of HTML table in a new Notebook cell ======================
        notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))


def _set_star_value(star_code, number_stars):
    """
    Internal function that is used for update the number of active stars (that define notebook
    difficulty level)

    ----------
    Parameters
    ----------
    star_code : str
        String with the HTML code to be changed.

    number_stars : int
        Number of stars that will be active.

    Returns
    -------
    out : str
        It is returned a string with the HTML code after updating the number of active stars.

    """

    for star in range(1, 6):
        if star <= number_stars:
            star_code = star_code.replace("fa fa-star " + str(star), "fa fa-star "
                                                                             "checked")
        else:
            star_code = star_code.replace("fa fa-star " + str(star), "fa fa-star")

    return star_code


def _generate_notebook_by_tag_body(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the page where notebooks are organized by
    tag values.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.

    """

    tag_keys = list(dict_by_tag.keys())
    tag_keys.sort()
    for tag in tag_keys:
        if tag.lower() not in SIGNAL_TYPE_LIST:
            markdown_cell = TAG_TABLE_HEADER
            markdown_cell = markdown_cell.replace("Tag i", tag)
            for notebook_file in dict_by_tag[tag]:
                split_path = notebook_file.replace("\\", "/").split("/")
                notebook_type = split_path[-2]
                notebook_name = split_path[-1].split("&")[0]
                notebook_title = split_path[-1].split("&")[1]
                markdown_cell += "\t<tr>\n\t\t<td width='20%' class='header_image_color_" + \
                                 str(NOTEBOOK_KEYS[notebook_type]) + "'><img " \
                                 "src='../../images/icons/" + notebook_type.title() +\
                                 ".png' width='15%'>\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='60%' class='center_cell open_cell_light'>" + \
                                 notebook_title + "\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='20%' class='center_cell'>\n\t\t\t<a href='" \
                                 "../" + notebook_type.title() + "/" + notebook_name + \
                                 "'><div class='file_icon'></div></a>\n\t\t</td>\n\t</tr>"

            markdown_cell += "</table>"

            # ==================== Insertion of HTML table in a new Notebook cell ======================
            notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))


def _generate_notebooks_by_category(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the page "Notebooks by Category".

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.

    """

    # ============================ Insertion of an opening text ====================================
    markdown_cell = OPEN_IMAGE

    # == Generation of a table that group Notebooks by category the information about each signal ==
    category_list = list(NOTEBOOK_KEYS.keys())
    tag_keys = list(dict_by_tag.keys())

    markdown_cell += """\n<table id="notebook_list" width="100%">
    <tr>
        <td width="20%" class="center_cell group_by_header_grey"> Category </td>
        <td width="60%" class="center_cell group_by_header"></td>
        <td width="20%" class="center_cell"></td>
    </tr>"""

    for i, category in enumerate(category_list):
        if category != "MainFiles":
            if category.lower() in tag_keys:
                if i == 0:
                    first_border = "color1_top"
                else:
                    first_border = ""


                nbr_notebooks = len(dict_by_tag[category.lower()])
                markdown_cell += "\n\t<tr>" \
                                 "\n\t\t<td rowspan='" + str(nbr_notebooks + 1) + "' class='center_cell open_cell_border_" + str(NOTEBOOK_KEYS[category]) + "'><span style='float:center'><img src='../../images/icons/" + category + ".png' class='icon' style='vertical-align:middle' alt='biosignalsnotebooks | " + category + " icon'></span> <span style='float:center' class='color" + str(NOTEBOOK_KEYS[category]) + "'>" + category + "</span></td>" \
                                 "\n\t\t<td class='center_cell color" + str(NOTEBOOK_KEYS[category]) + "_cell " + first_border + "'><span style='float:center'>" + category.replace("_", " ").replace("And", "and") + "</span></td>" \
                                 "\n\t\t<td class='center_cell gradient_color" + str(NOTEBOOK_KEYS[category]) + "'></td>" \
                                 "\n\t</tr>"

                notebook_list = dict_by_tag[category.lower()]
                for j, notebook_file in enumerate(notebook_list):
                    if j == len(notebook_list) - 1:
                        last_border = "class='border_cell_bottom_white'"
                    else:
                        last_border = ""

                    split_path = notebook_file.replace("\\", "/").split("/")
                    notebook_name = split_path[-1].split("&")[0]
                    notebook_title = split_path[-1].split("&")[1]
                    markdown_cell += "\n\t<tr " + last_border + ">" \
                                     "\n\t\t<td class='center_cell open_cell_light'> <a href='../" + category + "/" + notebook_name + "'>" + notebook_title + "</a> </td>" \
                                     "\n\t\t<td class='center_cell'> <a href='../" + category + "/" + notebook_name + "'><div class='file_icon'></div></a> </td>" \
                                     "\n\t</tr>"

    markdown_cell += "\n</table>"

    # ============================ Insertion of an introductory text ===============================
    markdown_cell += DESCRIPTION_CATEGORY

    # =================== Insertion of the HTML table inside a markdown cell =======================
    notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))


def _generate_github_readme(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the GitHub and PyPI README file.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.

    """

    # ==================== List of urls that contain each Category icon ============================
    icons = {"Detect": "https://i.ibb.co/rymrvFL/Detect.png",
             "Evaluate": "https://i.ibb.co/yfwcy2M/Evaluate.png",
             "Extract": "https://i.ibb.co/tchq7Cc/Extract.png",
             "Load": "https://i.ibb.co/YPbCnzD/Load.png",
             "Pre-Process": "https://i.ibb.co/1rKWccX/Pre-Process.png",
             "Record": "https://i.ibb.co/d2jZH1s/Record.png",
             "Train_And_Classify": "https://i.ibb.co/CQ4cyGb/Train-and-Classify.png",
             "Understand": "https://i.ibb.co/MnhRRQT/Understand.png",
             "Visualise": "https://i.ibb.co/wh4HKzf/Visualise.png",
             "Other": "https://i.ibb.co/ry9BzhV/Other.png",
             "Install": "https://i.ibb.co/4fBR4Q3/Install.png",
             "Connect": "https://i.ibb.co/3yDZpxC/Connect.png"}

    # =========================== biosignalsnotebooks website ======================================
    biosignalsnotebooks_web = "http://www.biosignalsplux.com/notebooks/Categories/"

    # == Generation of a table that group Notebooks by category the information about each signal ==
    category_list = list(NOTEBOOK_KEYS.keys())
    tag_keys = list(dict_by_tag.keys())

    markdown_cell = """<table width="100%">
    <tr>
        <td width="20%" align="center"><strong> Category <strong></td>
        <td width="80%"></td>
    </tr>"""

    for i, category in enumerate(category_list):
        if category != "MainFiles":
            if category.lower() in tag_keys:
                nbr_notebooks = len(dict_by_tag[category.lower()])
                notebook_list = dict_by_tag[category.lower()]
                split_path = notebook_list[0].replace("\\", "/").split("/")
                notebook_name = split_path[-1].split("&")[0]
                notebook_title = split_path[-1].split("&")[1]
                markdown_cell += "\n\t<tr>" \
                                 "\n\t\t<td rowspan='" + str(nbr_notebooks) + "'><p align='center'><img src='" + icons[category] + "' width='50%' align='center'></p></td>" \
                                 "\n\t\t<td align='center'> <a href='" + biosignalsnotebooks_web + category + "/" + notebook_name.replace(".ipynb", "_rev.php") + "' target='_blank'>" + notebook_title + "</a> </td>" \
                                 "\n\t</tr>"

                for j, notebook_file in enumerate(notebook_list[1:]):
                    split_path = notebook_file.replace("\\", "/").split("/")
                    notebook_name = split_path[-1].split("&")[0]
                    notebook_title = split_path[-1].split("&")[1]
                    markdown_cell += "\n\t<tr>" \
                                     "\n\t\t<td align='center'> <a href='" + biosignalsnotebooks_web + category + "/" + notebook_name.replace(".ipynb", "_rev.php") + "'>" + notebook_title + "</a> </td>" \
                                     "\n\t</tr>"

    markdown_cell += "\n</table>"

    # ============================ Generation of README files ======================================
    # [Open template]
    template_path = os.path.abspath(__file__).split(os.path.basename(__file__))[0].replace("\\", "/") + "/notebook_files/github/README_TEMPLATE.md"
    with open(template_path, 'r') as readme:
        readme_str = readme.read()

    readme_str = readme_str.replace("LIST_OF_NOTEBOOKS", markdown_cell)

    # [Storage of the updated files]
    for path in ["../biosignalsnotebooks/README_BSN.md", "../README.md"]:
        with open(path, 'w') as readme_out:
            readme_out.write(readme_str)
        readme_out.close()


def _generate_notebook_by_signal_type_body(notebook_object, dict_by_tag):
    """
    Internal function that is used for generation of the page where notebooks are organized by
    signal type where they are applicable.

    ----------
    Parameters
    ----------
    notebook_object : notebook object
        Object of "notebook" class where the body will be created.

    dict_by_tag : dict
        Dictionary where each key is a tag and the respective value will be a list containing the
        Notebooks (title and filename) that include this tag.
    """

    tag_keys = list(dict_by_tag.keys())
    tag_keys.sort()
    for tag in tag_keys:
        if tag.lower() in SIGNAL_TYPE_LIST:
            markdown_cell = TAG_TABLE_HEADER
            markdown_cell = markdown_cell.replace("Tag i", tag.upper())
            for notebook_file in dict_by_tag[tag]:
                split_path = notebook_file.replace("\\", "/").split("/")
                notebook_type = split_path[-2]
                notebook_name = split_path[-1].split("&")[0]
                notebook_title = split_path[-1].split("&")[1]
                markdown_cell += "\t<tr>\n\t\t<td width='20%' class='header_image_color_" + \
                                 str(NOTEBOOK_KEYS[notebook_type]) + "'><img " \
                                                                     "src='../../images/icons/" + notebook_type.title() + \
                                 ".png' width='15%'>\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='60%' class='center_cell open_cell_light'>" + \
                                 notebook_title + "\n\t\t</td>"
                markdown_cell += "\n\t\t<td width='20%' class='center_cell'>\n\t\t\t<a href='" \
                                 "../" + notebook_type.title() + "/" + notebook_name + \
                                 "'><div class='file_icon'></div></a>\n\t\t</td>\n\t</tr>"

            markdown_cell += "</table>"

            # ==================== Insertion of HTML table in a new Notebook cell ======================
            notebook_object["cells"].append(nb.v4.new_markdown_cell(markdown_cell))

def _generate_group_by_pages(root):
    dict_group_by_diff, dict_group_by_tag = _search_in_notebooks(root)
    file_path = root

    # Generation of biosignalsnotebooks environment main files.
    filename = "biosignalsnotebooks"
    main_page = notebook("Main_Files_By_Category", dict_by_difficulty=dict_group_by_diff,
                         dict_by_tag=dict_group_by_tag, notebook_file=filename)
    main_page.write_to_file(file_path, filename)
    filename = "by_diff"
    by_difficulty = notebook("Main_Files_By_Difficulty", "Notebooks Grouped by Difficulty",
                             dict_by_difficulty=dict_group_by_diff,
                             dict_by_tag=dict_group_by_tag,
                             notebook_file=filename)
    by_difficulty.write_to_file(file_path, filename)

    filename = "by_tag"
    by_tags = notebook("Main_Files_By_Tag", "Notebooks Grouped by Tag Values",
                       dict_by_difficulty=dict_group_by_diff, dict_by_tag=dict_group_by_tag,
                       notebook_file=filename)
    by_tags.write_to_file(file_path, filename)

    filename = "by_signal_type"
    by_signal_type = notebook("Main_Files_By_Signal_Type",
                              "Notebooks Grouped by Signal Type",
                              dict_by_difficulty=dict_group_by_diff,
                              dict_by_tag=dict_group_by_tag,
                              notebook_file=filename)
    by_signal_type.write_to_file(file_path, filename)

    filename = "signal_samples"
    signal_samples = notebook("Main_Files_Signal_Samples", "Signal Samples Library",
                              notebook_file=filename)
    signal_samples.write_to_file(file_path, filename)

def _search_in_notebooks(root):
    # Variable definition.
    dict_group_by_diff = {}
    dict_group_by_tag = {}

    categories = os.listdir(root + "/Categories")
    for category in categories:
        current_folder_path = root + "/Categories/" + category
        list_files = os.listdir(current_folder_path)
        for file in list_files:
            current_file_path = current_folder_path + "/" + file
            # Access to all Notebook files (with the extension .ipynb)
            if file.endswith(".ipynb"):
                # Read of the current Notebook.
                notebook = nb.read(current_file_path, nb.NO_CONVERT)

                # Search for "header" and/or "footer".
                header_cell, footer_cell, title, nbr_stars, tags = _get_metadata(notebook, file,
                                                                                 category)

                # Storage of Notebook metadata in global dictionaries.
                if category != "MainFiles":
                    if str(nbr_stars) not in (dict_group_by_diff.keys()):
                        dict_group_by_diff[str(nbr_stars)] = []
                    dict_group_by_diff[str(nbr_stars)].append(current_file_path + "&" + title)

                    for tag in tags:
                        if tag not in (dict_group_by_tag.keys()):
                            dict_group_by_tag[str(tag)] = []
                        dict_group_by_tag[str(tag)].append(current_file_path + "&" + title)

    return dict_group_by_diff, dict_group_by_tag


def _get_metadata(notebook, filename, category):
    # Variable initialisation.
    header_cell = None
    footer_cell = None
    title = None
    nbr_stars = None
    tags = None
    list_cells = notebook["cells"]

    # Search for "header" and "footer" tags and collect Notebook "difficulty" and "tags".
    for cell_nbr, cell in enumerate(list_cells):
        if "tags" in list(cell["metadata"].keys()):
            # [Header and Footer identification]
            if "header" in cell["metadata"]["tags"] and header_cell is None:
                header_cell = cell_nbr
            elif "footer" in cell["metadata"]["tags"] and footer_cell is None:
                footer_cell = cell_nbr
            elif "aux" in cell["metadata"]["tags"] and footer_cell is None:
                footer_cell = cell_nbr
            elif ("header" in cell["metadata"]["tags"] and header_cell is not None) or \
                    ("footer" in cell["metadata"]["tags"] and footer_cell is not None):
                raise RuntimeError("Duplicated 'header' or 'cell' tags inside the Notebook " +
                               filename + " !")

            # [Title, Difficulty and Tags]
            if category != "MainFiles":
                # [Title]
                if "intro_info_title" in cell["metadata"]["tags"]:
                    cell_content = cell["source"]

                    # Notebook Title.
                    title = cell_content.split('<td class="header_text">')[1].split('</td>')[0]

                # [Tags]
                if "intro_info_tags" in cell["metadata"]["tags"]:
                    cell_content = cell["source"]

                    # Notebook tag list.
                    tags = cell_content.split('<td class="shield_right" id="tags">')[1].split('</td>')[0].split("&#9729;")

                    # Difficulty level.
                    nbr_stars = cell_content.count("checked")

    return header_cell, footer_cell, title, nbr_stars, tags

# 29/11/2018  17h18m :)



"""
biosignalsnotebooks module responsible for loading data from the available signal samples (used on
biosignalsnotebooks Notebooks).

Available Functions
-------------------
[Public]

load_signal
    Function that returns a dictionary with the data contained inside a signal sample file
    (stored in the biosignalsnotebooks signal samples directory).

[Private]

_generate_download_google_link
    Google Drive sharable links do not allow a direct download with wget function, so, this private
    function manipulates the input link to ensure the desired download.

Observations/Comments
---------------------
None

/\
"""

# =================================================================================================
# ====================================== Import Statements ========================================
# =================================================================================================
import os
from .load import load
from .aux_functions import _generate_download_google_link

SIGNAL_PATH = (os.path.abspath(__file__).split(os.path.basename(__file__))[0] + \
              "notebook_files\\osf_files\\signal_samples\\").replace("\\", "/")
# TODO Substitute \\ by os.sep to work on all operative systems
FILE_EXTENSION = ".h5"


def load_signal(signal_handler, get_header=False):
    """
    -----
    Brief
    -----
    Function that returns a dictionary with the data contained inside 'signal_name' file (stored in
    the biosignalsnotebooks signal samples directory).

    -----------
    Description
    -----------
    Biosignalsnotebooks library provides data samples in order to the users that are new to biosignals data handling to
    have a place to start without the need to acquire new data. This sample files are stored in the folder
    _signal_samples inside the library.

    This function returns the data from the selected sample.

    ----------
    Parameters
    ----------
    signal_name : file name or url
        Name that identifies the signal sample to be loaded or a url.

        Possible values:
        [ecg_4000_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:12.4
        Sample Rate                4000 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

        [ecg_5_min]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           05:00.0
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

        [ecg_sample]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:11.9
        Sample Rate                 200 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

        [ecg_20_sec_10_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:20.0
        Sample Rate                  10 Hz
        Number of Channels               1
        Conditions                 At Rest
                             using Lead II
        =================   ==============

        [ecg_20_sec_100_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:19.7
        Sample Rate                 100 Hz
        Number of Channels               1
        Conditions                 At Rest
                             using Lead II
        =================   ==============

        [ecg_20_sec_1000_Hz]
        =================   ==============
        Signal Type                    ECG
        Acquisition Time           00:20.4
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions                 At Rest
                             using Lead II
        =================   ==============

        [emg_bursts]
        =================   ==============
        Signal Type                    EMG
        Muscle              Biceps Brachii
        Acquisition Time           00:28.5
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions             Cyclic
                               Contraction
        =================   ==============

        [emg_fatigue]
        =================   ==============
        Signal Type                    EMG
        Muscle              Biceps Brachii
        Acquisition Time           02:06.9
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions          Cyclic Flexion
                            and Extension
                            for fatigue
                            induction
        =================   ==============

        [temp_res_8_16]
        =================   ==============
        Signal Type            Temperature
        Acquisition Time           03:53.1
        Sample Rate                1000 Hz
        Number of Channels               2
        Resolutions          8 and 16 bits
        Conditions           Temperature
                             increase and
                             decrease
        =================   ==============

        [bvp_sample]
        =================   ==============
        Signal Type                    BVP
        Acquisition Time           00:27.3
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions                 At Rest
        =================   ==============

        [eeg_sample_closed_open_eyes]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           04:02.0
        Sample Rate                1000 Hz
        Number of Channels               2
        Conditions           Detection of
                             neuronal alpha
                             waves while
                             eyes are
                             opened or
                             closed
        =================   ==============

        [eeg_sample_artefacts_seg1]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           02:08.5
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           An EEG acqui-
                             sition which
                             contains
                             movement
                             artefacts
                             [Segment 1]
        =================   ==============

        [eeg_sample_artefacts_seg2]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           01:49.2
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           An EEG acqui-
                             sition which
                             contains
                             movement
                             artefacts
                             [Segment 2]
        =================   ==============

        [eeg_sample_artefacts_seg3]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           01:30.6
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           An EEG acqui-
                             sition which
                             contains
                             noisy data
                             with electro-
                             magnetic origin
                             Segment 3]
        =================   ==============

        [eeg_sample_0cm_h]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           03:59.7
        Sample Rate                1000 Hz
        Number of Channels               2
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed at the
                             Occipital line
                             connecting O1
                             and O2 positions
                             [Space between
                             electrodes = 0 cm]
        =================   ==============

        [eeg_sample_1cm_h]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           04:02.6
        Sample Rate                1000 Hz
        Number of Channels               2
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed at the
                             Occipital line
                             connecting O1
                             and O2 positions
                             [Space between
                             electrodes = 1 cm]
        =================   ==============

        [eeg_sample_2cm_h]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           04:17.9
        Sample Rate                1000 Hz
        Number of Channels               2
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed at the
                             Occipital line
                             connecting O1
                             and O2 positions
                             [Space between
                             electrodes = 2 cm]
        =================   ==============

        [eeg_sample_0cm_v]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           04:03.4
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed perpen-
                             dicularly to
                             the Occipital
                             line connecting
                             O1 and O2
                             positions
                             [Space between
                             electrodes = 0 cm]
        =================   ==============

        [eeg_sample_1cm_v]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           04:01.8
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed perpen-
                             dicularly to
                             the Occipital
                             line connecting
                             O1 and O2
                             positions
                             [Space between
                             electrodes = 1 cm]
        =================   ==============

        [eeg_sample_2cm_v]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           05:02.5
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed perpen-
                             dicularly to
                             the Occipital
                             line connecting
                             O1 and O2
                             positions
                             [Space between
                             electrodes = 2 cm]
        =================   ==============

        [eeg_sample_o]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           03:59.3
        Sample Rate                1000 Hz
        Number of Channels               2
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed at O1
                             and O2 positions
        =================   ==============

        [eeg_sample_p]
        =================   ==============
        Signal Type                    EEG
        Acquisition Time           04:02.3
        Sample Rate                1000 Hz
        Number of Channels               2
        Conditions           An EEG acquis-
                             ition with the
                             exploratory
                             electrodes
                             placed at P3
                             and P4 positions
        =================   ==============

        [emg_1000_hz_16_bits_solo]
        =================   ==============
        Signal Type                    EMG
        Acquisition Time           01:27.6
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           EMG acquisition
                             from Adductor
                             pollicis muscle
                             (1000 Hz and
                             16 bits)
        =================   ==============

        [emg_1000_hz_16_bits_solo]
        =================   ==============
        Signal Type                    EMG
        Acquisition Time           01:27.6
        Sample Rate                1000 Hz
        Number of Channels               1
        Conditions           EMG acquisition
                             from Adductor
                             pollicis muscle
                             (1000 Hz and
                             16 bits)
        =================   ==============

    get_header : boolean
        If True the file header will be returned as one of the function outputs.

    Returns
    -------
    out : dict
        A dictionary with the data stored inside the file specified in the input 'signal_name'.

    header : dict
        Metadata of the acquisition file (includes sampling rate, resolution, used device...)
    """

    primordial_list_signals = ["ecg_4000_Hz", "ecg_5_min", "ecg_sample", "ecg_20_sec_10_Hz",
                               "ecg_20_sec_100_Hz", "ecg_20_sec_1000_Hz", "emg_bursts", "emg_fatigue",
                               "temp_res_8_16", "bvp_sample", "eeg_sample_closed_open_eyes",
                               "eeg_sample_artefacts_seg1", "eeg_sample_artefacts_seg2", "eeg_sample_artefacts_seg3",
                               "eeg_sample_0cm_h", "eeg_sample_1cm_h", "eeg_sample_2cm_h", "eeg_sample_0cm_v",
                               "eeg_sample_1cm_v", "eeg_sample_2cm_v", "eeg_sample_o", "eeg_sample_p"]

    # Check if signal_handler is a url.
    # [Statements to be executed if signal_handler is a url]
    if any(mark in signal_handler for mark in ["http://", "https://", "www.", ".pt", ".com", ".org",
                                               ".net"]):
        # Check if it is a Google Drive sharable link.
        if "drive.google" in signal_handler:
            signal_handler = _generate_download_google_link(signal_handler)

        # Load file.
        out, header = load(signal_handler, remote=True, get_header=True, signal_sample=True)

    # [Statements to be executed if signal_handler is an identifier of the signal]
    else:
        list_files = list_signal_samples()
        if signal_handler in primordial_list_signals or signal_handler in list_files:
            out, header = load(SIGNAL_PATH + signal_handler + FILE_EXTENSION, get_header=True, signal_sample=True)
        else:
            raise RuntimeError("The signal name defined as input does not correspond to any of the "
                           "signal samples contained in the package.")

    if get_header is True:
        return out, header
    else:
        return out

def list_signal_samples():
    """
    -----
    Brief
    -----
    A "getter" that returns all signal samples filenames.

    -----------
    Description
    -----------
    The current function is a simpler "getter" returning the list of files inside SIGNALS_PATH with .h5 extension, i.e.,
    the files containing physiological data.

    Returns
    -------
    out : list
        A list containing all valid .h5 files.
    """

    # Get the complete list of files inside SIGNALS_PATH.
    list_files = os.listdir(SIGNAL_PATH)
    valid_files = []
    for file in list_files:
        if ".h5" in file:
            valid_files.append(file.split('.h5')[0])

    return valid_files

# 11/10/2018 16h45m :)



# coding=utf-8
"""
Functions intended to synchronise two signals and generate a new file containing the new synchronised signals.

Available Functions
-------------------
[Public]

synchronise_signals
    This function synchronises the input signals using the full cross correlation function between the signals.

generate_sync_txt_file
    This function allows to generate a text file with synchronised signals from the input file(s).

generate_sync_h5_file
    This function allows to generate a h5 file with synchronised signals from the input file(s).

create_synchronised_files
    This function creates .txt and .h5 files with synchronised signals.

Available Functions
-------------------
[Private]

_shape_array
    Function that equalises the input arrays by zero-padding the shortest one.

_create_h5_file_old
    Function to create a new .h5 file that contains the copy of the contents of the input file(s).

_create_h5_file
    Function to create a new .h5 file that contains the copy of the contents of the input file(s).

_create_txt_from_str
    This function allows to generate a text file with synchronised signals from the input file.

_create_txt_from_list
    This function allows to generate a text file with synchronised signals from the input files.

Observations/Comments
---------------------
None

/\
"""

import biosignalsnotebooks as bsnb
import numpy as np
from h5py import File


def synchronise_signals(in_signal_1, in_signal_2):
    """
    -----
    Brief
    -----
    This function synchronises the input signals using the full cross correlation function between the signals.

    -----------
    Description
    -----------
    Signals acquired with two devices may be dephased. It is possible to synchronise the two signals by multiple
    methods. Here, it is implemented a method that uses the calculus of the cross-correlation between those signals and
    identifies the correct instant of synchrony.

    This function synchronises the two input signals and returns the dephasing between them, and the resulting
    synchronised signals.

    ----------
    Parameters
    ----------
    in_signal_1 : list or numpy.array
        One of the input signals.
    in_signal_2 : list or numpy.array
        The other input signal.

    Returns
    -------
    phase : int
        The dephasing between signals in data points.
    result_signal_1: list or numpy.array
        The first signal synchronised.
    result_signal_2: list or numpy.array
        The second signal synchronised.
    """

    mean_1, std_1, mean_2, std_2 = [np.mean(in_signal_1), np.std(in_signal_1), np.mean(in_signal_2),
                                    np.std(in_signal_2)]
    signal_1 = in_signal_1 - mean_1
    signal_1 /= std_1
    signal_2 = in_signal_2 - mean_2
    signal_2 /= std_2

    # Calculate the full cross-correlation between the two signals.
    correlation = np.correlate(signal_1, signal_2, 'full')

    # Finding the edge of the correct correlation signal
    center = len(correlation) - len(signal_1) if len(signal_1) < len(signal_2) else len(correlation) - len(signal_2)

    # Finding the position of the maximum value of the correlation signal
    max_position = correlation.argmax()
    # Calculating the difference between the center and the position of the maximum value
    phase_straight = center - max_position

    # Finding the position of the maximum value of the reversed signal (corr[::-1])
    max_position_reversed = correlation[::-1].argmax()
    # Calculating the difference between the center and the position of the maximum value in the reversed correlation
    # signal
    phase_reversed = center - max_position_reversed

    # Calculate the dephasing between the signals. Maximum value of both results guarantees that we find the true
    # dephasing of the signals
    phases_aux = [phase_straight, phase_reversed]
    phase = np.abs(phases_aux).argmax()
    true_phase = np.abs(phases_aux[phase])

    if phases_aux[0] < phases_aux[1]:
        signal_1 = signal_1[true_phase:]
    else:
        signal_2 = signal_2[true_phase:]

    result_signal_1 = signal_1 * std_1 + mean_1
    result_signal_2 = signal_2 * std_2 + mean_2

    return true_phase, result_signal_1, result_signal_2


def generate_sync_txt_file(in_path, channels=("CH1", "CH1"), new_path='sync_file.txt'):
    """
    -----
    Brief
    -----
    This function allows to generate a text file with synchronised signals from the input file(s).

    -----------
    Description
    -----------
    OpenSignals files follow a specific structure that allows to analyse all files in the same way. Furthermore, it
    allows those files to be opened and analysed in the OpenSignals software without the need of programming.

    This functions takes one or two files, synchronises the signals in channels and generates a new file in the new
    path.

    ----------
    Parameters
    ----------
    in_path : str or list
        If the input is a string, it is assumed that the two signals are in the same file, else, if the input is a list,
        it is assumed that the two signals are in different file (the list should contain the paths to the two files).
    channels : list
        List with the strings identifying the channels of each signal. (default: ("CH1", "CH1"))
    new_path : str
        The path to create the new file. (default: 'sync_file.txt')
    """
    if type(in_path) is str:
        _create_txt_from_str(in_path, channels, new_path)
    elif type(in_path) is list:
        _create_txt_from_list(in_path, channels, new_path)
    else:
        raise TypeError('The path should be a list of str or a str.')


def generate_sync_h5_file(in_paths, channels=('channel_1', 'channel_1'), new_path='sync_file.h5'):
    """
    -----
    Brief
    -----
    This function allows to generate a h5 file with synchronised signals from the input file(s).

    -----------
    Description
    -----------
    OpenSignals files follow a specific structure that allows to analyse all files in the same way. Furthermore, it
    allows those files to be opened and analysed in the OpenSignals software without the need of programming.

    This functions takes one or two files, synchronises the signals in channels and generates a new file in the new
    path.

    ----------
    Parameters
    ----------
    in_paths : str or list
        If the input is a string, it is assumed that the two signals are in the same file, else, if the input is a list,
        it is assumed that the two signals are in different file (the list should contain the paths to the two files).
    channels : list
        List with the strings identifying the channels of each signal. (default: ("channel_1", "channel_1"))
    new_path : str
        The path to create the new file. (default: 'sync_file.h5')
    """

    if type(in_paths) == list or type(in_paths) == str:
        new_file = _create_h5_file(in_paths, new_path)
    else:
        raise TypeError('The path should be a list of str or a str.')

    data, devices = [[], []]
    for j, i in enumerate(list(new_file.keys())):
        devices.append(i)
        data.append(np.concatenate(new_file[i].get('raw').get(channels[j]))[:])
    phase, s1, s2 = synchronise_signals(data[0], data[1])

    if np.array_equal(s1, data[0]):
        # Change the second device
        new_device = new_file[devices[1]]
    elif np.array_equal(s2, data[1]):
        # Change the first device
        new_device = new_file[devices[0]]
    raw_digital = ['raw', 'digital', 'support']
    for r_d in raw_digital:
        raw_di = new_device[r_d]
        for key in list(raw_di.keys()):
            if r_d != 'support':
                signal = list(raw_di[key])[phase:]
                attribute = list(raw_di[key].attrs.items())
                del raw_di[key]
                raw_di.create_dataset(name=key, data=np.array(signal, dtype=np.int32))
                for a in attribute:
                    raw_di[key].attrs.__setitem__(name=a[0], value=a[1])
            else:
                for other_key in list(raw_di[key].keys()):
                    channel = raw_di[key][other_key]
                    level = int(key.split('_')[-1])
                    for yak in list(channel.keys()):
                        if yak is not 't':
                            data_aux = list(channel[yak])[int(np.ceil(phase/level)):]
                        else:
                            data_aux = list(channel[yak])[:len(channel[yak][:])-int(np.ceil(phase / level))]
                        del channel[yak]
                        channel.create_dataset(name=yak, data=np.array(data_aux))

    length = abs(len(new_file[devices[0]]['raw']['channel_1']) - len(new_file[devices[1]]['raw']['channel_1']))

    if len(new_file[devices[0]]['raw']['channel_1']) > len(new_file[devices[1]]['raw']['channel_1']):
        index = 1
    else:
        index = 0

    for r_d in raw_digital:
        raw_di = new_file[devices[index]][r_d]
        for key in list(raw_di.keys()):
            if r_d != 'support':
                if 'channel' in key or 'digital' in key:
                    signal = np.concatenate([list(raw_di[key]), np.zeros([length, 1])])
                elif 'nSeq' in key:
                    signal = np.arange(len(raw_di[key]) + length)
                attribute = list(raw_di[key].attrs.items())
                del raw_di[key]
                raw_di.create_dataset(name=key, data=np.array(signal, dtype=np.int32))
                for a in attribute:
                    raw_di[key].attrs.__setitem__(name=a[0], value=a[1])
            else:
                for other_key in list(raw_di[key].keys()):
                    channel = raw_di[key][other_key]
                    level = int(key.split('_')[-1])
                    for yak in list(channel.keys()):
                        if yak is not 't':
                            data_aux = np.vstack([list(channel[yak]),
                                                  np.zeros(int(length / level)).reshape(-1, 1)])
                        else:
                            data_aux = np.arange(0, len(channel[yak][:])+length//level).reshape(-1, 1)*level

                        del channel[yak]
                        channel.create_dataset(name=yak, data=np.array(data_aux, dtype=int))

    final_length = len(np.ravel(signal))
    for i in list(new_file.keys()):
        new_file[i].attrs['duration'] = str(int(final_length // new_file[i].attrs['sampling rate'])) + ' s'
        new_file[i].attrs.modify(name='nsamples', value=int(final_length))
    new_file.close()


def create_synchronised_files(in_path=(('file1.txt', 'file2.txt'), ('file1.h5', 'file2.h5')), channels=(1, 1),
                      file_name='sync_file'):
    """
    -----
    Brief
    -----
    This function creates .txt and .h5 files with synchronised signals.

    -----------
    Description
    -----------
    OpenSignals software generates 3 types of files structured in specific ways. The formats are .txt, .h5 and .edf.
    Those files follow specific structures in order to normalise their analysis and to be able to be opened by the
    OpenSignals software.

    This functions allows to generate .txt and .h5 files with synchronised signals structured the same way as the files
    generated by OpenSignals software.

    ----------
    Parameters
    ----------
    in_path : list
        List containing the paths to the two file formats. If the signals are in separate files, each entry should be a
        list of strings, else, the list should contain two strings.
        (default: (('file1.txt', 'file2.txt'), ('file1.h5', 'file2.h5') - case in which the two signals are in separate
        files)
        (example in which the signals are in the same file: ('file.txt', 'file.h5')
    channels : list
        List with the ints identifying the channels of each signal. (default: (1, 1))
    file_name : str
        The name of the new files without the extension. (default: 'sync_file')
    """
    h5_channels = ['channel_'+str(i) for i in channels]
    txt_channels = ['CH' + str(i) for i in channels]
    generate_sync_h5_file(in_path[1], channels=h5_channels, new_path=file_name+'.h5')
    generate_sync_txt_file(in_path[0], channels=txt_channels, new_path=file_name + '.txt')


# ==================================================================================================
# ================================= Private Functions ==============================================
# ==================================================================================================


def _shape_array(array1, array2):
    """
    Function that equalises the input arrays by zero-padding the shortest one.

    ----------
    Parameters
    ----------
    array1: list or numpy.array
        Array
    array2: list or numpy.array
        Array

    Return
    ------
    arrays: numpy.array
        Array containing the equal-length arrays.
    """
    if len(array1) > len(array2):
        new_array = array2
        old_array = array1
    else:
        new_array = array1
        old_array = array2

    length = len(old_array) - len(new_array)

    for i in range(length):
        n = new_array[-1].copy()
        n[0::3] += 1
        n[2::3] = 0
        new_array = np.vstack([new_array, [n]])

    arrays = np.hstack([old_array, new_array])
    return arrays


def _create_h5_file_old(in_paths, new_path):
    """
    Function to create a new .h5 file that contains the copy of the contents of the input file(s).

    in_paths : str or list
        If the input is a string, it is assumed that the two signals are in the same file, else, if the input is a list,
        it is assumed that the two signals are in different file (the list should contain the paths to the two files).
    new_path : str
        The path to create the new file. (default: 'sync_file.h5')

    Returns
    -------
    new_file : h5py Object
        Object of the h5py package containing the new file containing the copy of the contents of the input file(s).
    """
    if type(in_paths) == str:
        in_paths = [in_paths]
    new_file = File(new_path, 'w')
    for in_path in in_paths:
        with File(in_path) as file:
            for attribute in (list(file.attrs.items())):
                new_file.attrs.__setitem__(name=attribute[0], value=attribute[1])
            for key in list(file.keys()):
                new_file.create_group(key)
                for attribute in list(file[key].attrs.items()):
                    new_file[key].attrs.__setitem__(name=attribute[0], value=attribute[1])
                for other_key in list(file[key].keys()):
                    new_file.create_group(key + '/' + other_key)
                    for attribute in list(file[key + '/' + other_key].attrs.items()):
                        new_file[key + '/' + other_key].attrs.__setitem__(name=attribute[0], value=attribute[1])
                    for another_key in list(file[key][other_key]):
                        try:
                            for yet_another_key in list(file[key][other_key][another_key].keys()):
                                try:
                                    for y in list(file[key][other_key][another_key][yet_another_key].keys()):
                                        new_file.create_dataset(name=key + '/' + other_key + '/' + another_key + '/'
                                                                     + yet_another_key + '/' + y,
                                                                data=file[key][other_key][another_key][yet_another_key][
                                                                    y])
                                        for attribute in list(
                                                file[key][other_key][another_key][yet_another_key].attrs.items()):
                                            new_file[key + '/' + other_key + '/' + another_key + '/'
                                                     + yet_another_key + '/' + y].attrs.__setitem__(
                                                name=attribute[0],
                                                value=attribute[1])
                                except:
                                    new_file.create_dataset(
                                        name=key + '/' + other_key + '/' + another_key + '/' + yet_another_key,
                                        data=file[key][other_key][another_key][yet_another_key])
                                    for attribute in list(
                                            file[
                                                key + '/' + other_key + '/' + another_key + '/' + yet_another_key].attrs.items()):
                                        new_file[
                                            key + '/' + other_key + '/' + another_key + '/' + yet_another_key].attrs.__setitem__(
                                            name=attribute[0],
                                            value=attribute[1])
                        except:
                            new_file.create_dataset(name=key + '/' + other_key + '/' + another_key,
                                                    data=list(file[key][other_key][another_key]))
                            for attribute in list(file[key + '/' + other_key + '/' + another_key].attrs.items()):
                                new_file[key + '/' + other_key + '/' + another_key].attrs.__setitem__(name=attribute[0],
                                                                                                      value=attribute[
                                                                                                          1])

    return new_file


def _create_h5_file(in_paths, new_path):
    """
    Function to create a new .h5 file that contains the copy of the contents of the input file(s).

    in_paths : str or list
        If the input is a string, it is assumed that the two signals are in the same file, else, if the input is a list,
        it is assumed that the two signals are in different file (the list should contain the paths to the two files).
    new_path : str
        The path to create the new file. (default: 'sync_file.h5')

    Returns
    -------
    new_file : h5py Object
        Object of the h5py package containing the new file containing the copy of the contents of the input file(s).
    """
    if type(in_paths) == str:
        in_paths = [in_paths]
    new_file = File(new_path, 'w')
    for i, in_path in enumerate(in_paths):
        with File(in_path, 'r') as file:
            for key in list(file.keys()):
                file.copy(source=file[key], dest=new_file, name=key)

    return new_file


def _create_txt_from_str(in_path, channels, new_path):
    """
    This function allows to generate a text file with synchronised signals from the input file.

    ----------
    Parameters
    ----------
    in_path : str
        Path to the file containing the two signals that will be synchronised.
    channels : list
        List with the strings identifying the channels of each signal.
    new_path : str
        The path to create the new file.
    """
    header = ["# OpenSignals Text File Format"]
    files = [bsnb.load(in_path)]
    with open(in_path, encoding="latin-1") as opened_p:
        header.append(opened_p.readlines()[1])
    header.append("# EndOfHeader")

    data = []
    nr_channels = []
    for file in files:
        for i, device in enumerate(file.keys()):
            nr_channels.append(len(list(file[device])))
            data.append(file[device][channels[i]])

    dephase, s1, s2 = synchronise_signals(data[0], data[1])

    new_header = [h.replace("\n", "") for h in header]
    sync_file = open(new_path, 'w')
    sync_file.write(' \n'.join(new_header) + '\n')

    old_columns = np.loadtxt(in_path)
    if np.array_equal(s1, data[0]):
        # Change the second device
        aux = 3 * nr_channels[0]
        columns = old_columns[dephase:, aux:]
        new_file = _shape_array(old_columns[:, :aux], columns)
    elif np.array_equal(s2, data[1]):
        # Change the first device
        aux = 3 * nr_channels[1]
        columns = old_columns[dephase:, :aux]
        new_file = _shape_array(columns, old_columns[:, aux:])
    else:
        print("The devices are synchronised.")
        return
    for line in new_file:
        sync_file.write('\t'.join(str(int(i)) for i in line) + '\t\n')
    sync_file.close()


def _create_txt_from_list(in_path, channels, new_path):
    """
    This function allows to generate a text file with synchronised signals from the input files.

    ----------
    Parameters
    ----------
    in_path : list
        Paths to the files containing the two signals that will be synchronised.
    channels : list
        List with the strings identifying the channels of each signal.
    new_path : str
        Path to create the new file.
    """
    header = ["# OpenSignals Text File Format"]
    files = [bsnb.load(p) for p in in_path]
    with open(in_path[0], encoding="latin-1") as opened_p:
        with open(in_path[1], encoding="latin-1") as opened_p_1:
            header.append(opened_p.readlines()[1][:-2] + ', ' + opened_p_1.readlines()[1][3:])
    header.append("# EndOfHeader")

    data = []
    nr_channels = []
    for i, file in enumerate(files):
        nr_channels.append(len(list(file)))
        data.append(file[channels[i]])

    dephase, s1, s2 = synchronise_signals(data[0], data[1])

    new_header = [h.replace("\n", "") for h in header]
    sync_file = open(new_path, 'w')
    sync_file.write('\n'.join(new_header) + '\n')

    if np.array_equal(s1, data[0]):
        # Change the second device
        old_columns = np.loadtxt(in_path[1])
        columns = old_columns[dephase:]
        new_file = _shape_array(columns, np.loadtxt(in_path[0]))
    elif np.array_equal(s2, data[1]):
        # Change the first device
        old_columns = np.loadtxt(in_path[0])
        columns = old_columns[dephase:]
        new_file = _shape_array(columns, np.loadtxt(in_path[1]))
    else:
        print("The devices are synchronised.")
        return
    for line in new_file:
        sync_file.write('\t'.join(str(int(i)) for i in line) + '\t\n')
    sync_file.close()
